;--------------------------------------------------------------------------
; X64 example that fills a buffer and prints some msg on the
; screen.
; -------------------------------------------------------------------------

SIZE_OF_BUFFER	EQU 32*1024*1024

EXTERN ExitProcess:PROC
EXTERN GlobalAlloc:PROC
EXTERN Sleep:PROC
EXTERN printf:PROC
EXTERN _getch:PROC
EXTERN GetCurrentProcess:PROC
EXTERN SetProcessAffinityMask:PROC

INCLUDELIB Kernel32.lib
INCLUDELIB msvcrt.lib


.DATA

    msg  BYTE "Clearing done",13,10,0
    msg1 BYTE "%u clocks for a %u bytes buffer with using %s",13,10,13,10,0
    repstosq$	db	"REP STOSQ",0
    movntdq$	db	"MOVNTDQ",0

.DATA?

    buf2clear DQ	?

.CODE

align 16
main PROC

    push rbx	; align stack
    
    mov rcx,40h	; GMEM_ZEROINIT
    add rsp,-32	; for params
    mov rdx,SIZE_OF_BUFFER
    call GlobalAlloc
    
    mov [buf2clear],rax
    
    call GetCurrentProcess
    xchg rax,rcx
    mov rdx,1
    call SetProcessAffinityMask
    
    ; commit as much as possible allowable for testing machine
    mov rdx,[buf2clear]
    mov rcx, SIZE_OF_BUFFER
    call ClearBufferSSE2
    
    ; First test - REP STOSQ
    
    mov rcx,300	; 300 msec
    call Sleep
    
    cpuid
    rdtsc
    shl rdx,32
    lea rbx,[rdx+rax]
        
    mov rdx,[buf2clear]
    mov rcx, SIZE_OF_BUFFER
    neg rbx	; for later addition instead substraction
    call ClearBuffer
    
    cpuid
    rdtsc
    shl rdx,32
    add rbx,rax
    add rbx,rdx
    
    lea rcx, msg
    call printf

    lea r9,repstosq$
    mov r8,SIZE_OF_BUFFER
    mov rdx,rbx
    lea rcx, msg1
    call printf
    
    
    ; Second test - SSE2 MOVNTDQ
    
    mov rcx,300	; 300 msec
    call Sleep
    
    cpuid 
    rdtsc
    shl rdx,32
    lea rbx,[rdx+rax]
    
    ;note: buffer must be 16bytes aligned, but for GlobalAlloc this is ok.    
    mov rdx,[buf2clear]
    mov rcx, SIZE_OF_BUFFER
    neg rbx	; for later addition instead substraction
    call ClearBufferSSE2
    
    cpuid
    rdtsc
    shl rdx,32
    add rbx,rax
    add rbx,rdx
    
    lea rcx, msg
    call printf
    
    lea r9,movntdq$
    mov r8,SIZE_OF_BUFFER
    mov rdx,rbx
    lea rcx, msg1
    call printf 
    
    

    call _getch	; wait for any key, don't display
    
    ; as good programmers we must restore stack and free the buffer
    ; but this is not needed due to ExitProcess.
    ; so no needed this lines:
    ;add rsp,32
    ;pop rbx

    xor rcx, rcx
    call ExitProcess
    int 3

main ENDP

; -------------------------------------------------------------------------
; This PROC clear a buffer
; -------------------------------------------------------------------------
; void ClearBuffer(ULONGLONG sizeOfBufferInBytes, PVOID addressOfBuffer);
align 16
ClearBuffer PROC

    xchg rdx,rdi	; pointer in RDX, number of bytes in RCX already
    shr rcx,3	; RCX = RCX/8, because input is in bytes, not QWORDs   

    mov rax, 2020202020202020h

    rep stosq

    mov rdi,rdx

    ret

ClearBuffer ENDP


; -------------------------------------------------------------------------
; This PROC clear a buffer, SSE2 version
; -------------------------------------------------------------------------
; void ClearBufferSSE2(ULONGLONG sizeOfBufferInBytes, PVOID addressOfBuffer);
align 16
ClearBufferSSE2 PROC

    mov eax, 20202020h
    movd xmm0,eax
    pshufd xmm0,xmm0,0

    align 16
    ; used Frank's variation of SSE algo
    @@:
    movntdq [rdx],xmm0
    add rdx,16
    add rcx,-16
    jg @B   

    ret

ClearBufferSSE2 ENDP

align 16
WaitKey PROC
	jmp _getch	; simple thunk to _getch, WaitKey leaved for compatibility reasons :)
WaitKey ENDP

END main